/* Copyright (C) 2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * If the integration of miniscope code requires the use of a license
 * other than GPL, an alternative license may be arranged.
 */

#include "HidScope.h"
#include "..\miniscope\device\MiniscopeDevice.h"
#include "HID\HidDevice.h"
#include "HID\DeviceNotifier.h"
#include <windows.h>
#include <stdio.h>
#include <assert.h>

using namespace nsHidDevice;
HidDevice hidDevice;


int CallbackDeviceState(std::string path, bool attached)
{
    if (stricmp(path.c_str(), hidDevice.GetPath().c_str()))
        return 0;
    if (!attached)
    {
        hidScope->ConnectCb(DEVICE_DISCONNECTED, "Device was detached!");
        hidScope->log( "Device detached." );
        hidScope->setOpened(false);
    }
    return 0;
}

nsDeviceNotifier::DeviceNotifier deviceNotifier;


DWORD WINAPI ThreadTimerProc(LPVOID data)
{
    while (1)
    {
        Sleep(hidScope->sampling_period);
        SetEvent ( hidScope->hEvent );
    }
}

DWORD WINAPI ThreadProc(LPVOID data)
{
    while (1)
    {
        WaitForSingleObject ( hidScope->hEvent, INFINITE );
        if ((int)hidScope->sample_count >= hidScope->sampling_frame || !hidScope->bDeviceOpened)
            continue;

        if (hidScope->sample_count == 0)
        {
            if (hidScope->callbackTrigger)
                hidScope->callbackTrigger();
            hidScope->samples.clear();
        }

        // get single sample from device
        char sendbuf[64];
        char rcvbuf[64];

        sendbuf[0] = 0x37;  //request code (see PIC firmware)
        int status = hidDevice.WriteReport(HidDevice::E_REPORT_OUT, 0, sendbuf, sizeof(sendbuf));
        if (status == 0)
        {
            int len = sizeof(rcvbuf);
            status = hidDevice.ReadReport(HidDevice::E_REPORT_IN, 0, rcvbuf, &len, 500);
            if (status == 0)
            {
                short val = (unsigned char)rcvbuf[1];        //LSB
                val += 256 * (unsigned char)rcvbuf[2]; //MSB
                //hidScope->log("rcv val = %d", val);
                hidScope->samples.push_back(val);
                if (hidScope->samples.size() == 64)
                {
                    if (hidScope->callbackData)
                        hidScope->callbackData(&hidScope->samples[0], hidScope->samples.size() * sizeof(hidScope->samples[0]));
                    hidScope->samples.clear();
                }
                hidScope->sample_count++;
            }
        }
    }
}


HidScope::HidScope():
        sampling_period(100),
        sampling_frame(512),
        Thread(NULL),
        ThreadTimer(NULL)
{
    bDeviceOpened = false;
    sample_count = sampling_frame;
    bStopRequest = false;
    callbackData = NULL;
    callbackLog = NULL;
    callbackConnect = NULL;

    OutputDebugStr("HidScope created");
    hEvent = CreateEvent ( NULL, false, false, "MyEvent" );
    assert(hEvent);
}

HidScope::~HidScope()
{
    Disconnect();
    CloseHandle(hEvent);
}

void HidScope::log(char *lpData, ...)
{
    static char buf[256]; //determines max message length
    va_list ap;
    va_start(ap, lpData);
    vsnprintf(buf, sizeof(buf), lpData, ap);
    va_end(ap);
    buf[sizeof(buf)-1] = 0;

    if (callbackLog)
        callbackLog(buf);
}

int HidScope::Connect(void)
{
    DWORD dwtid;
    if (!Thread)
        Thread = CreateThread(NULL, 0, ThreadProc, this, 0, &dwtid);
    if (!ThreadTimer)
        ThreadTimer = CreateThread(NULL, 0, ThreadTimerProc, this, 0, &dwtid);
    SetThreadPriority(Thread, THREAD_PRIORITY_HIGHEST);
    SetThreadPriority(ThreadTimer, THREAD_PRIORITY_TIME_CRITICAL);

    if (bDeviceOpened)
        hidDevice.Close();
    bDeviceOpened = false;
    int status = hidDevice.Open(VENDOR_ID, PRODUCT_ID, NULL, NULL);
    if (status)
    {
        ConnectCb(DEVICE_DISCONNECTED, "Could not connect to device. Check if device is connected. Restart is required.");
        log( "Device open failed with status = %d.", status );
        return 1;
    }
    else
    {
        log ("Device connected!");

        nsDeviceNotifier::DeviceNotifier::pCallbackFunction = &CallbackDeviceState;
        GUID hidGuid;
        hidDevice.GetHidGuid(&hidGuid);
        status = deviceNotifier.RegisterForNotifications(hidGuid);
        if (status)
        {
            //DisplayLastError("GetLastError: ");
            log("Failed to register for HID device removal notifications! Status = %d", status);
        }

        ConnectCb(DEVICE_CONNECTED, "PIC HID scope");
        bDeviceOpened = true;
        return 0;
    }
}

int HidScope::Disconnect(void)
{
    if (bDeviceOpened)
        hidDevice.Close();

    if (Thread)
        TerminateThread(Thread, 0);
    if (ThreadTimer)
        TerminateThread(ThreadTimer, 0);
    Thread = NULL;
    ThreadTimer = NULL;
    Sleep(10);

    bDeviceOpened = false;
    return 0;
}

int HidScope::ManualTrigger(void)
{
    //log("ManualTrigger");
    if (!bDeviceOpened)
    {
        log("ManualTrigger: dev not opened");
        return -1;
    }
    sample_count = 0;
    return 0;
}
